% -- External Function: [hmod, mask] = dry_puddles( h_bo );
%
% Take as input a two-dimensional array `h_bo' (BOttom depth) with values >=0,
%   where h_bo=0 is condidered `land' and h_bo>0 is considered `water'.
% Identify the different basins and attribute them a unique identification
%   number, then return the result as `mask'. Discard all the basins inside h_bo
%   except the largest one, and return the result as `hmod' (h MODified).
%
% Function arguments:
% `h_bo' (input ): Two-dimensional array of type single or double (described
%                    above).
% `hmod' (output): Two-dimensional array (same dimensions as h_bo) representing
%                    the bottom depth after discarding the `puddles'.
% `mask' (output): Two-dimensional array (same dimensions as h_bo) where each
%                    basin is identified with a unique identification number.
function [hmod, mask] = dry_puddles( h_bo )

% `EPSIlon', the smallest bottom depth considered. Cells of bottom depth
%   greater than epsi are considered `water', otherwise, `land'.
  epsi = 1.e-3; % (meters)

% The operations below assume that there is a 1 point-wide margin of zeros
%   surrounding the input array. If that is not the case, add it at this point
%   and remove it later.
  if ( any( [ h_bo(  :, 1  );  ...
              h_bo(  :, end);  ...
              h_bo(  1, :  )'; ...
              h_bo(end, :  )' ] ~= 0 ) )
    addm = 1;
    wrk1 = zeros( size( h_bo, 1 ) + 2, ...
                  size( h_bo, 2 ) + 2 );
    wrk1( 2 : end - 1, 2 : end - 1 ) = h_bo;
    h_bo = wrk1; clear wrk1;
  else
    addm = 0;
  end

  hmod = h_bo; % The copy of `h' that is being MODified.
  lm   = size( h_bo, 1 ) - 2;
  mm   = size( h_bo, 2 ) - 2;
  mask = zeros(    lm + 2, mm + 2  );
  neig = zeros( 4, lm + 2, mm + 2  ); % The 4 neighbors of a given i,j cell.

% Set wet grid cells that are not yet associated with a given basin to -1.
  mask( find( hmod > epsi ) ) = - 1;

% Pick a random point that is known to be wet, associate it with basin #1,
%   then use it as your starting point for the exploration of basin #1.
  nbas = 1; % Number of BASins identified so far.
  ix   = find( mask == - 1 );
  mask( ix( round(rand * length(ix)) ) ) = nbas; clear ix;

  iter = 0;
% As long as there are unindentifed wet cells, keep exploring.
  while ( any( mask(:) == - 1 ) )
    iter = iter + 1;
    neig(1, 2 : end - 1, 2 : end - 1) = mask(1 : end - 2, 2 : end - 1);
    neig(2, 2 : end - 1, 2 : end - 1) = mask(3 : end,     2 : end - 1);
    neig(3, 2 : end - 1, 2 : end - 1) = mask(2 : end - 1, 1 : end - 2);
    neig(4, 2 : end - 1, 2 : end - 1) = mask(2 : end - 1, 3 : end    );
    maxn = max( reshape( neig, 4, (lm + 2) * (mm + 2) ) );
    maxn = reshape( maxn, lm + 2, mm + 2 );

%   Get list of grid cells that are still unassociated and positioned next to
%     a grid cell that is part of the basin being currently explored (nbas).
    ix   = find( mask == - 1 & maxn == nbas );

    if ( length( ix ) > 0 ) % If there are such grid cells, register them.
      [ivec, jvec] = find( mask == - 1 & maxn == nbas );
      for ipnt = 1 : length( ivec )
        i    = ivec(ipnt);
        j    = jvec(ipnt);

%       Rather than registering one point at a time, accelerate things by
%         expanding as much as possible in the i and j directions.
        bad  = find( mask(:, j)  ~= - 1 & mask(:, j)  ~= nbas );
        imin = max( bad( find(bad < i) ) ) + 1;
        imax = min( bad( find(bad > i) ) ) - 1;

        bad  = find( mask(i, :)' ~= - 1 & mask(i, :)' ~= nbas );
        jmin = max( bad( find(bad < j) ) ) + 1;
        jmax = min( bad( find(bad > j) ) ) - 1;

        mask( imin : imax, j ) = nbas;
        mask( i, jmin : jmax ) = nbas;
      end; clear ipnt;
    else % All the cells of the basin being explored have been registered.
%     Note the area of this basin:
      area( nbas, 1 ) = length( find( mask == nbas ) );

%     Move on to the next basin to be explored:
      ix   = find( mask == - 1 );
      mask( ix( max([ 1; round(rand * length(ix)) ]) ) ) = nbas + 1;
      nbas = nbas + 1;
    end
    clear ix;
  end

% Note the area of the last basin that was explored:
  area( nbas, 1 ) = length( find( mask == nbas ) );

% Discard all the basins in array `hmod' except the largest one.
  for ibas = 1 : nbas
    if ( area( ibas ) < max( area ) )
      hmod( find( mask == ibas ) ) = 0.;
    end
  end; clear ibas;

  if ( addm ) % Remove the extra margins that were added earlier.
    hmod = hmod(2 : end - 1, 2 : end - 1);
    mask = mask(2 : end - 1, 2 : end - 1);
  end
end
