
% Test-case of a realistic topography and realistic wind stress field.
% The wind stress and topography need to be downloaded from public online
%   datasets (see below). The test-case also relies on a cartographic software
%   (see below) to generate a conformal map projection. The test-case represents
%   the Southern Ocean with an arbitrary horizontal mesh size `dl'.

  clear;
  close all;
  more  off;

  dl   = 50.e3;   % Mesh size (m).
% dl   = 10.e3;   % Mesh size (m).
  nlay = 3;       % Number of LAYers.
  dt_o = 5.;      % Frequency of outputs (days).
  rmax = 6000.e3; % MAXimum Radius of the circumpolar domain (m).
% Path to input/output files (must end with character '/'):
  outd = '/tmp/';
% Seawater potential density and vertical position for the TOP of the Layers
%   (`topl' will be non-dimensionalized later in the script). Profile is based
%   on World Ocean Atlas 2018, zonally-averaged at 53deg.S. This is the latitude
%   where Ekman transport switches from divergent to convergent hence isopycnals
%   will remain steady at that latitude.
  if     ( nlay == 2 )                 % 1st baroclinic Rossby radius = 11.7km.
    rhon = [ 1027.1,         1027.8 ]; % kg/m3.
    topl = [     0.,           250. ]; % m.
  elseif ( nlay == 3 )                 % 1st baroclinic Rossby radius = 14.2km.
    rhon = [ 1027.1, 1027.5, 1027.8 ]; % kg/m3.
    topl = [     0.,   250.,  1000. ]; % m.
  else
    error('Case not implemented yet.');
  end

% Define a Cartesian model domain centered around the South Pole. Inverse the
%   x,y positions into longitudes and latitudes using the software `proj'
%   (proj.org) which is readily available in many Linux/UNIX distributions. Any
%   conformal map projection is acceptable for beom, but we will use a
%   stereographic projection since it has the nice property of preserving
%   circular topographic shapes.

  disp('Obtaining the latitudes/longitudes of the projection...');
  lat0  = - 90.; % (degrees of latitude;  South Pole.        )
  lon0  =    0.; % (degrees of longitude; Greenwich meridian.)
  x     = (- rmax : dl : rmax);
  y     = (- rmax : dl : rmax);

  [yy, xx] = meshgrid( y, x ); clear x  y;
  lm    = size( yy, 1 ) - 2; % Grid size in x direction.
  mm    = size( yy, 2 ) - 2; % Grid size in y direction.
  xypo  = [ xx(:), yy(:) ];
  save -ascii '/tmp/xypos.txt' xypo;

  comd  = [ 'invproj +proj=stere +lon_0=' num2str( lon0 ) ...
            ' +lat_0=' num2str( lat0 ) ' -v -f "%.5f" '   ...
            ' < /tmp/xypos.txt > /tmp/lonlat.txt' ];
  system( comd ); clear lon0 lat0 comd;
  lola = load( '/tmp/lonlat.txt' );
  glon = reshape( lola(:, 1), [lm + 2, mm + 2] );
  glat = reshape( lola(:, 2), [lm + 2, mm + 2] );
  clear lola;

% At this point, the longitudes/latitudes of the Cartesian grid are defined and
%   can be used to extract the topography / wind stress from external datasets:
% Topography:  https://doi.org/10.7289/V5C8276M
% Wind stress: https://climate.copernicus.eu/climate-reanalysis
% N.B. These are not necessarily the best datasets available, but they are well
%   suited for a test-case (simple to use).

% Path to file containing the seabed/ice surface dataset:
  fice = '/media/pierre/data/scientif_data/bathymetry/etopo1/ETOPO1_Ice_g_gmt4.grd';
  xice = ncread( fice, 'x' ); % Longitudes of dataset.
  yice = ncread( fice, 'y' ); % Latitudes  of dataset.
  zice = ncread( fice, 'z' ); % Seabed/surface of continental ice (m).
  clear fice;

% Bi-linear interpolation of the topography onto the model grid:
  disp('Interpolating the topography...');
  h_bo = interp2( xice, yice, zice', glon, glat );
  clear xice yice zice;

  h_bo = h_bo * (- 1.); % BOttom depth of model grid, positive downward (m).

% The spatial interpolation sometimes fails at the South Pole.
  i__0 = floor(0.5 * (lm + 2)) + 1;
  j__0 = floor(0.5 * (mm + 2)) + 1;
  if ( isnan( h_bo( i__0, j__0 ) ) )
    h_bo( i__0, j__0 ) = 0.;
  end
  disp([ 'The domain runs from ' num2str(max(- glat(:, j__0))) 'S to ' ...
                                 num2str(min(- glat(:, j__0))) 'S.' ]);

% Keep note of the location of `land cells', here defined as depth < than the
%   thickness of the top layer. In a configuration with nlay > 2, this helps to
%   ensure that the influence of Salmon's term is limited to the baroclinic
%   modes of motion, aka `simple outcropping' in beom's documentation. All grid
%   cells satisfying this definition of `land' will be set to zero depth later
%   on. (Alternatively, one could dredge cells with 0<depth<topl(2) to enforce
%   a minimum depth = topl(2); this alternative strategy would have the benefit
%   of preserving the location of the coastline.)
  land = find( h_bo < topl(2) );

% The original topographic dataset contains small-scale (O(1km)) topography that
%   will never be resolved properly by the `dl' parameter selected above
%   (aliasing), contributing to abrupt transitions between `active' and
%   `outcropped' regions. A few iterations of a diffusion operator over the
%   topographic grid help lower the topographic slope, |grad(h_bo)|.
  cfl  = 0.25;    % Courant-Friedrichs-Lewy = nu * dt / dl^2, must be < 0.5.
  alph = 1. / 3.; % Lynch 1992, https://docs.lib.purdue.edu/cstech/929
  coef = cfl * [ 0.5 * alph,   1. - alph,              0.5 * alph; ...
                 1.  - alph, - 4. * (1. - 0.5 * alph), 1.  - alph; ...
                 0.5 * alph,   1. - alph,              0.5 * alph ];
  clear cfl alph;
  if     ( dl > 55.e3 ) % More smoothing when more roughness is introduced.
    nitr = 1;
  elseif ( dl > 35.e3 )
    nitr = 2;
  elseif ( dl > 15.e3 )
    nitr = 4;
  elseif ( dl >  5.e3 )
    nitr = 5;
  end
  for iter = 1 : nitr
    hsmo = h_bo; % H SMOothed.
    hsmo(    2 : end - 1, 2 : end - 1)                    ...
      = hsmo(2 : end - 1, 2 : end - 1) * (coef(2, 2) + 1) ...
      + hsmo(1 : end - 2, 2 : end - 1) *  coef(1, 2)      ...
      + hsmo(3 : end,     2 : end - 1) *  coef(3, 2)      ...
      + hsmo(2 : end - 1, 1 : end - 2) *  coef(2, 1)      ...
      + hsmo(2 : end - 1, 3 : end    ) *  coef(2, 3)      ...
      + hsmo(1 : end - 2, 1 : end - 2) *  coef(1, 1)      ...
      + hsmo(1 : end - 2, 3 : end    ) *  coef(1, 3)      ...
      + hsmo(3 : end,     1 : end - 2) *  coef(3, 1)      ...
      + hsmo(3 : end,     3 : end    ) *  coef(3, 3);
    hsmo(1,   :  ) = hsmo(2,       :);
    hsmo(end, :  ) = hsmo(end - 1, :);
    hsmo(:,   1  ) = hsmo(:,       2);
    hsmo(:,   end) = hsmo(:, end - 1);
    h_bo = hsmo;
  end; clear nitr iter hsmo;

% Compute the topographic slope. This information is used to guess-estimate the
%   appropriate `hsal'.
  dhdx = zeros( lm + 2, mm + 2 );
  dhdy = zeros( lm + 2, mm + 2 );
  dhdx( 2 : end, : ) = ( h_bo(2 : end, :) - h_bo(1 : end - 1, :) ) / dl;
  dhdy( :, 2 : end ) = ( h_bo(:, 2 : end) - h_bo(:, 1 : end - 1) ) / dl;
  slop = reshape( max([ abs(dhdx(:))'; abs(dhdy(:))' ])', [lm + 2, mm + 2] );
  clear dhdx dhdy;

% Salmon's scheme requires the transition between `outcropped' regions and
%   active regions to not be excessively abrupt, which is achieved by selecting
%   a sufficiently thick `hsal' value. An initial guess for hsal is obtained
%   from a rule of thumb valid in cases of `simple outcropping' (see beom's
%   documentation), but this guess should always be validated by running the
%   code without forcing (no winds) and verifying that currents remain neglible
%   over time. If growing spurious currents are apparent, then hsal should be
%   increased. Note that the rule of thumb is proportional to the mesh size,
%   meaning that smaller `dl' allow for much thinner hsal values (assuming the
%   topography's roughness isn't increased in parallel).
% hsal > (slope x dl) x (Delta rho / rho_0) ~ (slope x dl) x 1. / rho
  hsal = 13. * max( slop(:) ) * dl * 1. / rhon(end);

% In `simple outcropping', the top layer is assumed to be sufficiently thick
%   that the artifical Salmon term of the top layer plays a negligible role in
%   the state of rest of the top layer:
  if ( topl(2) < hsal * 10. )
    error([ 'The Salmon term of the top layer will not be negligible in '    ...
            'the state of rest; Increase the thickness of the top layer, '   ...
            'or decrease the topographic slopes, or decrease the mesh size ' ...
            '`dl`.' ]);
  end

% Discard grid cells positioned on land, and cells too shallow relative to your
%   guidelines defined above:
  h_bo( land                    ) = 0.; clear land;
  h_bo( find( h_bo <= topl(2) ) ) = 0.;

% Discard `puddles', i.e. lakes or any secondary basins that are not connected
%   to the main basin. This operation relies on a function stored in the file
%   `dry_puddles.m'.
  disp('Identifying basins over the model domain...');
  h_bo = dry_puddles( h_bo );

% Enforce a closed model domain (for the sake of simplicity) by defining dry
%   margins.
  h_bo( :,   1 ) = 0.;
  h_bo( :, end ) = 0.;
  h_bo(   1, : ) = 0.;
  h_bo( end, : ) = 0.;

% Discard grid cells beyond the `rmax' radius; this mimics a circumpolar
%   cylindrical domain and it lowers the total number of grid cells (faster).
% The result also resembles the zonally-reentrant channel of Ward & Hogg 2011
%   ( https:/doi.org/10.1016/j.ocemod.2011.08.004 ).
  h_bo( find( sqrt(xx.^2 + yy.^2) > rmax ) ) = 0.;

  figure; colormap( gray ); % Visualize the final bathymetry.
    hand = imagesc( h_bo' ); axis xy; axis equal; colorbar;
    title( 'Topography (meters)' );

% Now that the topography is finalized, get the number of degrees of freedom
%   and non-dimensionalize `topl' by the maximum depth of the basin:
  ndeg = get_nbr_deg_freedom( h_bo );
  topl = topl / max( h_bo(:) );       % Non-dimensional.

% Path to file containing the wind stress:
  fstr = '/media/pierre/data/scientif_data/era5/global_monthly_wind_stress2013_2022.nc';
  o_er = ncread( fstr, 'longitude' );
  a_er = ncread( fstr, 'latitude'  );
  iews = ncread( fstr, 'iews'      ); % Zonal      stress (Pa <=> N m^(-2)).
  inss = ncread( fstr, 'inss'      ); % Meridional stress (Pa <=> N m^(-2)).
  lasm = ncread( fstr, 'lsm'       ); % LAnd(1)/Sea(0) Mask of this dataset.
  iews = squeeze( mean( iews, 3 )  ); % Average over time (3rd dimension.)
  inss = squeeze( mean( inss, 3 )  );
  lasm = squeeze( mean( lasm, 3 )  );

% There is no guarantee that the land/sea boundary of this wind dataset matches
%   exactly the one from your topographic model grid, and we don't want land
%   wind values to leak over the ocean. Set wind stress on land to NaN (Not a
%   Number) and then later on extrapolate over the missing values.
  iews( find( lasm > 0.01 ) ) = nan;
  inss( find( lasm > 0.01 ) ) = nan;
  clear lasm;

% This particular dataset requires 0<long<360 rather than -90<long<90:
  long = glon;
  long( find( long < 0. ) ) = long( find( long < 0. ) ) + 360.;

  disp('Interpolating wind stress...');
  iews = interp2( o_er, a_er, iews', long, glat );
  inss = interp2( o_er, a_er, inss', long, glat );
  clear o_er a_er long;

% Extrapolate over cells with undefined values (NaN), based on the 4 neighbors.
%   You do this after the interpolation in case there is a NaN near the pole.
  while( any( isnan( iews(:) ) ) )
    neig = repmat( iews, [1, 1, 4] );
    neig(2 : end,     :, 1) = iews(1 : end - 1, :);
    neig(1 : end - 1, :, 2) = iews(2 : end,     :);
    neig(:, 2 : end,     3) = iews(:, 1 : end - 1);
    neig(:, 1 : end - 1, 4) = iews(:, 2 : end    );
    neig = squeeze( nanmean( permute( neig, [3, 1, 2] ) ) );
    iews( find( isnan( iews ) ) ) = neig( find( isnan( iews ) ) );
    clear neig;
  end
  while( any( isnan( inss(:) ) ) )
    neig = repmat( inss, [1, 1, 4] );
    neig(2 : end,     :, 1) = inss(1 : end - 1, :);
    neig(1 : end - 1, :, 2) = inss(2 : end,     :);
    neig(:, 2 : end,     3) = inss(:, 1 : end - 1);
    neig(:, 1 : end - 1, 4) = inss(:, 2 : end    );
    neig = squeeze( nanmean( permute( neig, [3, 1, 2] ) ) );
    inss( find( isnan( inss ) ) ) = neig( find( isnan( inss ) ) );
    clear neig;
  end

% The code expects the wind stress to be aligned with the axes of the grid.
% Therefore, the zonal and meridional stress need to be rotated. First,
%   determine the angle for the rotation (`angl'):
  dodx = nan( lm + 2, mm + 2 ); % dlongitude/dx.
  dody = nan( lm + 2, mm + 2 ); % dlongitude/dy.
  dodx(2 : end - 1, :) = (glon(3 : end, :) -  glon(1 : end - 2, :)) / 2.;
  dody(:, 2 : end - 1) = (glon(:, 3 : end) -  glon(:, 1 : end - 2)) / 2.;
  dodx(  1, :) = glon(  2, :) - glon(      1, :);
  dodx(end, :) = glon(end, :) - glon(end - 1, :);
  dody(:,   1) = glon(:,   2) - glon(:,       1);
  dody(:, end) = glon(:, end) - glon(:, end - 1);

% dodx,dody have discontinuities where the longitude jumps from -180deg. to
%   +180deg and next to the South Pole. Set them to NaN (Not a Number).
  dodx( find( abs(dodx) > 90. ) ) = nan;
  dody( find( abs(dody) > 90. ) ) = nan;
  angl = atan2( dody, dodx ); clear dodx dody;

% Extrapolate over cells with undefined values (NaN), based on the 4 neighbors.
  while( any( isnan( angl(:) ) ) )
    neig = repmat( angl, [1, 1, 4] );
    neig(2 : end,     :, 1) = angl(1 : end - 1, :);
    neig(1 : end - 1, :, 2) = angl(2 : end,     :);
    neig(:, 2 : end,     3) = angl(:, 1 : end - 1);
    neig(:, 1 : end - 1, 4) = angl(:, 2 : end    );
    neig = squeeze( nanmean( permute( neig, [3, 1, 2] ) ) );
    angl( find( isnan( angl ) ) ) = neig( find( isnan( angl ) ) );
    clear neig;
  end

% At this point the rotation angle is known over the whole model domain.
% Rotation = [cos(angl), - sin(angl)]
%            [sin(angl),   cos(angl)]
  taux = iews .* cos( angl ) - inss .* sin( angl );
  tauy = iews .* sin( angl ) + inss .* cos( angl );

% Recall that we have closed boundaries, meaning that we will have coastal
%   downwelling at the boundary (not very realistic) unless we taper
%   off the winds in the vicinity of the closed boundary. (This also mimics
%   the analytical wind stress of Ward & Hogg 2011.)
  x53s = abs( glat(:, j__0) + 53. ); x53s = find( x53s == min( x53s ) );
  x53s = (x53s( end ) - i__0) * dl;
  radi = (sqrt( xx.^2 + yy.^2 ) - x53s) / (xx(end, j__0) - x53s) * pi / 2.;
  radi( find( radi < 0.      ) ) = 0.;
  radi( find( radi > pi / 2. ) ) = pi / 2.;
  tapr = cos( radi ); clear x53s radi;

  taux = taux .* tapr;
  tauy = tauy .* tapr;
  iews = iews .* tapr; % iews,inss only used for visualization at this point.
  inss = inss .* tapr; clear tapr;

% Illustrate the zonally-averaged zonal wind stress and topography:
  laax = (min(glat(:)) + 1. : 1. : max(glat(:, j__0)))';
  zawt = nan( length( laax ), 1 );
  for ilat = 1 : length( laax )
    indx = find( glat >= laax(ilat) - 1. ...
               & glat <  laax(ilat) );
    zawt( ilat, 1 ) = mean( iews( indx ) );
    zawt( ilat, 2 ) = mean( h_bo( indx ) ); clear indx;
  end; clear ilat;
  figure;
  subplot(2, 1, 1);
    hand = plot( laax, zawt(:, 1), [laax(1), laax(end)], [0., 0.] );
    set( hand, 'linewidth', 3 );
    title('Zonally-averaged zonal wind stress and topography');
    ylabel('Zonal wind stress (Pascals)');
  subplot(2, 1, 2);
    hand = plot( laax, zawt(:, 2) ); axis ij;
    set( hand, 'linewidth', 3 );
    ylabel('Depth (m)');
    xlabel('Latitude (deg.)');
  clear iews inss laax zawt hand;

% Compute the curl of the wind stress, which defines locations of upwelling or
%   downwelling:
  curl = zeros( size( taux, 1 ), size( taux, 2 ) );
  curl( 2 : end - 1, 2 : end - 1) ...
    = (tauy(3 : end, 2 : end - 1) - tauy(1 : end - 2, 2 : end - 1)) /(2.*dl) ...
    - (taux(2 : end - 1, 3 : end) - taux(2 : end - 1, 1 : end - 2)) /(2.*dl);
  curl( find( h_bo < 1.e-3 ) ) = 0.;

  figure; colormap( jet ); % Visualize curl:
    hand = imagesc( curl' ); axis xy; axis equal;
    title('Curl of wind stress (Pa m^{-1})');
    caxis([- 2.e-6, 2.e-6]); colorbar; clear curl;

% CORiolis parameter `f':
  fcor = 2. * 2. * pi / 24. / 3600. * sind( glat );

% Celerity of EXTernal gravity waves (determines time-step):
  grav = 9.8;                           % GRAVity (m/s2).
  cext = sqrt( grav * max( h_bo(:) ) ); % (m/s).

% `hmin' is the minimum layer thickness allowed during a calculation.
%   By default, the code assumes `hsal = 10 x hmin':
  hmin = hsal / 10.; % Meters.

% Note that a 2-day ramp (dt_r=2.) is applied on the wind forcing to prevent
%   the inertial oscillations that accompany a sudden increase in winds
%   (e.g., Millot and Crepon 1981).
% print_params( lm,   mm,   nlay, ndeg, dl,   cext, f0,   rhon, topl, dt_s, ...
%               dt_o, dt_r, dt3d, bvis, dvis, bdrg, hmin, hsbl, hbbl, g_fb, ...
%               uadv, qdrg, ocrp, rsta, xper, yper, diag, tauw, idir, odir, ...
%               desc );
  print_params( lm,   mm,   nlay, ndeg, dl,   cext, 0.,   rhon, topl, 30.,  ...
                dt_o, 2.,   0.1,  0.,   0.1,  3.e-3,hmin, 100., 100., 1.,   ...
                1.,   1.,   1.,   0,    0.,   0.,   1.,   [0,0],outd, outd, ...
                'Test-case: Southern Ocean with realistic topo and winds' );

% Write input files over disk. Before doing so, make sure that the directory
%   `outd' is clean (i.e., delete all input files inside `outd').

  if ( exist([ outd 'init.bin' ]) ); delete([ outd 'init.bin' ]); end
  if ( exist([ outd 'h_bo.bin' ]) ); delete([ outd 'h_bo.bin' ]); end
  if ( exist([ outd 'fcor.bin' ]) ); delete([ outd 'fcor.bin' ]); end
  if ( exist([ outd 'taus.bin' ]) ); delete([ outd 'taus.bin' ]); end
  if ( exist([ outd 'bodf.bin' ]) ); delete([ outd 'bodf.bin' ]); end
  if ( exist([ outd 'nudg.bin' ]) ); delete([ outd 'nudg.bin' ]); end
  if ( exist([ outd 'hdot.bin' ]) ); delete([ outd 'hdot.bin' ]); end
  if ( exist([ outd 'tide.bin' ]) ); delete([ outd 'tide.bin' ]); end

  [f_id, mesg] = fopen([ outd 'h_bo.bin' ],         'w', 'ieee-le' );
  cont = fwrite( f_id, h_bo,               'real*4', 0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg cont;

  [f_id, mesg] = fopen([ outd 'fcor.bin' ],         'w', 'ieee-le' );
  cont = fwrite( f_id, fcor,               'real*4', 0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg cont;

  [f_id, mesg] = fopen([ outd 'taus.bin' ],         'w', 'ieee-le' );
  cont = fwrite( f_id, cat(3, taux, tauy), 'real*4', 0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg cont;

% See beom documentation on how to compile and execute the program.

  disp(['Now compile and execute the program in a different terminal. ' ...
        'Press a key when it is over.']);
  pause

  disp('Reading the output data, please wait...');

% Read the calculation's metadata:
  [taxi, h_0, f0, dl, rhon, desc] = get_metadata( strtrim( outd ) );
  nrec = length( taxi ); % Number of RECords in the output files.

% Visualize the sea surface height and surface velocity. Technically, the `u'
%   and `v' components of velocity are not co-located and we ideally would
%   interpolate them at a common location (the center of the grid cell) prior
%   to visualizing, but this is just a test-case.
  figure; colormap( jet );
  for irec = 1 : nrec
    n    = get_field( 'eta_', irec, outd );
    u    = get_field( 'u___', irec, outd );
    v    = get_field( 'v___', irec, outd );
    subplot(1, 2, 1);
      imagesc( squeeze( n(:,:,1) )' );
      title([ 'Sea surface elevation (m), Day='     num2str(irec * dt_o) ]);
      colorbar; axis xy; axis equal;
    subplot(1, 2, 2);
      imagesc( squeeze( sqrt( u(:,:,1).^2 + v(:,:,1).^2 ) )' );
      title([ 'Velocity top layer (m s^{-1}), Day=' num2str(irec * dt_o) ]);
      colorbar; axis xy; axis equal;
    drawnow;
  end; clear irec;

